__all__ = ['PX', 'PY', 'PZ', 'PM', 'PP', 'VPZ', 'RPY', 'ID', 'HD',
           'CZ', 'iSWAP', 'gate_expand_2toN', 'isVirtual', 'isPhysicalGate',
           'isTwoQubitGate']

"""
Backend with definition of 1qb and 2qb gates
"""

from . import qubitClass as Qb
from qutip import qeye, create, destroy, tensor, Qobj, basis, sigmax, sigmay
from numpy import pi, exp

def ID(Qblist, target=None):
    """Creates identity gate which can be used when creating a openQASM file to make
    sure all gates are in the correct depth"""
    id = [qeye(Qb.level) for Qb in Qblist]
    return tensor(id)

def PX(Qblist, target):
    """Creates specific sigmax gate.
    Input:
    - Qblist = list of Qubit objects
    - target = qubit to be targeted by the operator
    Output: Qobj that operates on qubit[target] with correct dimensions"""
    sx = [qeye(Qb.level) for Qb in Qblist]
    sx[target] = destroy(Qblist[target].level) + create(Qblist[target].level)
    return tensor(sx)

def PY(Qblist, target):
    """Creates specific sigmay gate.
    Input:
    - Qblist = list of Qubit objects
    - target = qubit to be targeted by the operator
    Output: Qobj that operates on qubit[target] with correct dimensions"""
    sy = [qeye(Qb.level) for Qb in Qblist]
    sy[target] = -1j * (destroy(Qblist[target].level) - create(Qblist[target].level))
    return tensor(sy)

def RPY(Qblist, target):
    """REVERSED SIGMAY, TO BE USED IN HADAMARD
    Input:
    - Qblist = list of Qubit objects
    - target = qubit to be targeted by the operator
    Output: Qobj that operates on qubit[target] with correct dimensions"""
    rsy = [qeye(Qb.level) for Qb in Qblist]
    rsy[target] = 1j* (destroy(Qblist[target].level) - create(Qblist[target].level))
    # OBS: In HD we use -PY for negative rotation.
    return tensor(rsy)

def PM(Qblist, target):
    """Creates specific sigma- gate
    Input:
    - Qblist = list of Qubit objects
    - target = qubit to be targeted by the operator
    Output: Qobj that operates on qubit[target] with correct dimensions"""
    sm = [qeye(Qb.level) for Qb in Qblist]
    sm[target] = destroy(Qblist[target].level)
    return tensor(sm)

def PP(Qblist, target):
    """Creates specific sigma+ gate
    Input:
    - Qblist = list of Qubit objects
    - target = qubit to be targeted by the operator
    Output: Qobj that operates on qubit[target] with correct dimensions"""
    sp = [qeye(Qb.level) for Qb in Qblist]
    sp[target] = create(Qblist[target].level)
    return tensor(sm)

def PZ(Qblist, target):
    """Creates specific sigmaz gate
    Input:
    - Qblist = list of Qubit objects
    - target = qubit to be targeted by the operator
    Output: Qobj that operates on qubit[target] with correct dimensions"""
    sz = [qeye(Qb.level) for Qb in Qblist]
    sz[target] = 2 * create(Qblist[target].level) * destroy(Qblist[target].level) - qeye(Qblist[target].level)
    return tensor(sz)

def VPZ(Qblist, target, angle):
    """Creates virtual sigmaz gate,
    Input:
    - Qblist = list of Qubit objects
    - target = qubit to be targeted by the operator
    - angle = angle which to virtually rotate around z-axis
    Output: Qobj that operates on qubit[target] with correct dimensions and angle"""
    vsz = [qeye(Qb.level) for Qb in Qblist]
    try:
        if Qblist[target].level == 2:
            vsz[target] = Qobj([[exp(-1j*angle/2), 0], [0, exp(1j*angle/2)]])
        elif Qblist[target].level == 3:
            vsz[target] = Qobj([[exp(-1j * angle / 2), 0, 0], [0, exp(1j * angle / 2), 0], [0, 0, 1]])
        elif Qblist[target].level == 4:
            vsz[target] = Qobj([[exp(-1j * angle / 2), 0, 0, 0], [0, exp(1j * angle / 2), 0, 0], [0, 0, 1, 0], [0, 0, 0, 1]])
    except:
        print("Couldn't create Virtual Pauli Z gate! Maybe you dont have qubit level between 2 and 4?")
        print("Calculates the virtual pauli z gate as identity matrix!")
    return tensor(vsz)


def HD(Qblist, target):
    """Create Hadamard gate
    NOTE: Angle for HD_real is always pi/2 and for HD_virt always pi
    Input:
    - Qblist = list of Qubit objects
    - target = qubit to be targeted by the operator
    Output: Qobj that operates on qubit[target] with correct dimensions"""
    HD_real = RPY(Qblist, target)  # OBS negative PY rotation, that's why we call RPY
    HD_virt = VPZ(Qblist, target, pi)
    return [HD_real, HD_virt]

def CZ(Qblist, Tar_Con):
    """Create specific CZ gate, modelled as H = |11><02|+|02><11| or |11><20|+|20><11|
    NOTE: Only works for 3level qubits.
    Input:
    - Qblist = list of Qubit objects
    - Tar_Con = qubits to be targeted by the operator
    Output: Qobj that operates on the targeted qubits with correct dimensions"""
    k11 = tensor(basis(3, 1), basis(3, 1))
    k02 = tensor(basis(3, 0), basis(3, 2))
    H = k11 * k02.dag() + k02 * k11.dag()
    cz = [qeye(Qb.level) for Qb in Qblist]
    target = Tar_Con[0]  # index of the targeted qubit
    control = Tar_Con[1]  # index of the controlling qubit
    del(cz[max(Tar_Con)])  # Make room for the cz gate
    del(cz[min(Tar_Con)])  # Make room for the cz gate
    return gate_expand_2toN(H,len(Qblist),cz,control,target)

def iSWAP(Qblist, Tar_Con):
    """
    Saved this for now, might delete later.
    Since this is a symmetric swap, both qubits are targets and controls
    To avoid enumeration of the tar/con qubits, they are still called "target" and "control"
    Takes help from the gate_expand_2toN function
    Quantum object representing the iSWAP gate.
    Returns
    -------
    iSWAP_gate : qobj
        Quantum object representation of iSWAP gate

    target = Tar_Con[0]  # index of the targeted qubit
    control = Tar_Con[1]  # index of the controlling qubit
    tarLevel = Qblist[target].level
    conLevel = Qblist[control].level
    k01 = tensor(basis(conLevel, 0), basis(tarLevel, 1))
    k10 = tensor(basis(conLevel, 1), basis(tarLevel, 0))
    H = 1j * (k01 * k10.dag() + k10 * k01.dag())
    #We must add diagonal ones in order to treat all possible states (right???)
    size = H.shape[0] #H can be represented as an size x size matrix
    H2 = np.zeros([size, size])
    for i in range(size):
        H2[i,i] = np.array_equal(H[i],H[i] * 0) #if this row is all zero, we have to put a 1 at position (i,i) to "do nothing"
    H2 = Qobj(H2, dims = H.dims)
    #H = H + H2
    iSWAP = [qeye(Qb.level) for Qb in Qblist]
    del (iSWAP[max(Tar_Con)])  # Make room for the iSWAP gate
    del (iSWAP[min(Tar_Con)])  # Make room for the iSWAP gate
    return gate_expand_2toN(H, len(Qblist), iSWAP, control, target)"""
    raise Exception ("iSWAP is not implemented. See gateLib for more info")



def gate_expand_2toN(U, N, cz, control=None, target=None, targets=None):
    """Modified QuTiP function.
    Original source code: https://qutip.org/docs/latest/modules/qutip/qip/operations/gates.html"""
    """
    Create a Qobj representing a two-qubit gate that act on a system with N
    qubits.
    Parameters
    ----------
    U : Qobj
        The two-qubit gate
    N : integer
        The number of qubits in the target space.
    control : integer
        The index of the control qubit.
    target : integer
        The index of the target qubit.
    targets : list
        List of target qubits.
    Returns
    -------
    gate : qobj
        Quantum object representation of N-qubit gate.
    """

    if targets is not None:
        control, target = targets

    if control is None or target is None:
        raise ValueError("Specify value of control and target")

    if N < 2:
        raise ValueError("integer N must be larger or equal to 2")

    if control >= N or target >= N:
        raise ValueError("control and not target must be integer < integer N")

    if control == target:
        raise ValueError("target and not control cannot be equal")

    p = list(range(N))

    if target == 0 and control == 1:
        p[control], p[target] = p[target], p[control]

    elif target == 0:
        p[1], p[target] = p[target], p[1]
        p[1], p[control] = p[control], p[1]

    else:
        p[1], p[target] = p[target], p[1]
        p[0], p[control] = p[control], p[0]
    return tensor([U] + cz).permute(p)

"""Classification of gates"""
def isVirtual(gate, i):
    if gate.name[i] in ['VPZ']:
        return True
    else:
        return False

def isTwoQubitGate(gate, i):
    if gate.name[i] in ['CZ', 'iSWAP','CZ']:
        return True
    else:
        return False

def isPhysicalGate(gate,i):
    if gate.name[i] in ['PX', 'PY', 'PZ', 'PM', 'PP', 'ID']:
        return True
    else:
        return False

if __name__ == "__main__":
    """ Troubleshooting"""

    # Test specific sigmax
    sx1 = tensor(qeye(2), sigmax(), qeye(2))
    Qblist = [Qb.Qubit(2, [], [], [],[]) for i in range(0,3)]
    sx = PX(Qblist,1)
    print(sx)
    if sx1 == sx:
        print("Specific sigx works!")
    else:
        print("Specific sigx doesn't work")
        print(sx1)
        print(sx)

    # Test specific sigmay
    sy1 = tensor(qeye(2), sigmay(), qeye(2))
    sy = PY(Qblist,1)
    if sy1 == sy:
        print("Specific sigy works!")
    else:
        print("Specific sigy doesn't work")
        print(sy1)
        print(sy)

    # Test specific sigmam
    sm1 = tensor(qeye(2), destroy(2), qeye(2))
    sm = PM(Qblist, 1)
    if sy1 == sy:
        print("Specific sigm works!")
    else:
        print("Specific sigm doesn't work")
        print(sy1)
        print(sy)


    # Test specific sigmaz
    # But not the same as inbuilt sigmaz()?
    sz1 = sm1.dag()*sm1
    sz = PZ(Qblist,1)
    if sz1 == sz:
        print("Specific sigz works!")
    else:
        print("Specific sigz doesn't work")
        print(sz1)
        print(sz)

    # Test iSWAP
    Qblist = [Qb.Qubit(2, [], [], [],[]) for i in range(2)] + [Qb.Qubit(2, [], [], [],[])]
    iSWAP = iSWAP(Qblist, [0,2])
    print("iSWAP: ", iSWAP)

    print(sx*iSWAP)

    # Test CZ
    Qblist = [Qb.Qubit(3, [], [], [], []) for i in range(2)] + [Qb.Qubit(3, [], [], [], [])]
    CZnew = CZ(Qblist, [0, 1])
    q1 = tensor(basis(3,0), basis(3,2)) # the state |02>
    q2 = tensor(basis(3,1), basis(3,1)) # the state |11>
    print("CZ: ", CZnew)



__all__ = ['createCollapseOperators']

from math import sqrt
from . import gateLib as gl


def createCollapseOperators(Qblist):
    """Function for creating the collapse operators for mcsolve.
     Input: Qblist = List of Qubit objects (see Qb_class)
     output: list of tensored qobj with correct dimensions for each noise"""
    c_ops = []
    for i in range(0, len(Qblist)):
        if Qblist[i].noisert_vec[0] > 0.0:  # Relaxation
            c_ops.append(sqrt(Qblist[i].noisert_vec[0]) * gl.PM(Qblist, i))
        if Qblist[i].noisert_vec[1] > 0.0:  # Dephasing
            c_ops.append(sqrt(Qblist[i].noisert_vec[1]) * gl.PZ(Qblist, i)/2)
        if Qblist[i].noisert_vec[2] > 0.0:  # Thermal
            c_ops.append(sqrt(Qblist[i].noisert_vec[3]) * gl.PP(Qblist, i))
    return c_ops



__all__ = ['readFile']

from pandas import read_csv
from numpy import pi
from sys import exit

def readFile(filename, n):
    """
    Reads the data in the file
    Input:
    - filename = name of file to be read
    - n = number of qubits (rows of paramaters in the file to be read)
    Output: data in the file
    (Qubit data: relax, depha, therma, anharm, levels)
    """
    relax = []
    depha = []
    therma = []
    anharm = []
    levels = []
    try:
        data = read_csv(filename, sep=";")
        arr = data.to_numpy()
        for i in range(0, n):
            relax.append(arr[i, 1])
            depha.append(arr[i, 2])
            therma.append(arr[i, 3])
            anharm.append(arr[i, 4]*2*pi*1e6)  # Transfer from [MHz] to angular frequency
            levels.append(int(arr[i, 5]))
    except Exception as error:
        print("Something is wrong with the csv file! Check the example-file in git for how it should look."
              " Remember to use correct separation of values")
        raise exit(1)
    return relax, depha, therma, anharm, levels





__all__ = ['mcs', 'mcsTimeDynamics', 'virtgate']

from qutip import mcsolve, sesolve

def mcs(psi,H,tlist,c_ops):
    """
    Used for the parfor operations in mainAlgorithm.py.
    Inputs:
    - psi = state (qobj) which is iterated
    - H = Hamiltonian for the step
    - tlist = time list for the step
    - c_ops = collapse operators
    Output: Final state after each step
    """
    if c_ops != []:
        output = mcsolve(H, psi, tlist, c_ops=c_ops, e_ops = [], ntraj=1, progress_bar=None)
        outstate = (output.states[:, -1])
        return outstate[0]

    else:
        output = sesolve(H, psi, tlist, e_ops = [])
        outstate = output.states[-1]
        return outstate

def mcsTimeDynamics(psi, H, tlist, c_ops):
    """Use this if StoreTimeDynamics=True.
    Returns all states at each time in tlist instead of just final state"""
    if c_ops != []:
        output = mcsolve(H, psi, tlist, c_ops=c_ops, e_ops=[], ntraj=1, progress_bar=None)
        outstates = output.states
        return outstates
    else:
        output = sesolve(H, psi, tlist, e_ops=[])
        outstates = output.states
        return outstates


def virtgate(psi,vgate):
    """Applies virtual gate through matrix multiplication
    Input:
    - psi = State (Qobj ket)
    - vgate = virtual gate (Qobj matrix)
    Output: final state after virtual gate"""
    return vgate*psi


__all__ = ['qasmToQnas']

from qiskit.converters import circuit_to_dag, dag_to_circuit
from re import search
from numpy import pi
from . import gateFuncs as gf


def qasmToQnas(circuit):
    """
    Translates qiskit circuits to QnAS circuit.
    Input:
    - OpenQASM quantum circuit (qiskit QuantumCircuit object)
    - show_circuit = option to print circuit, optional
    Output: list of AlgStep objects for the algorithm.
    """
    qc = circuit
    qc.remove_final_measurements()  # Removing final measure

    def piDelivery(s):  # takes in string s and returns float of string
        """
        Translates strings to float
        Input:
        - string
        Output: float
        """
        flt = 0
        if '*pi/' in s:
            flt = float(s.split('*')[0]) * pi / float(s.split('/')[1])
        elif '*pi' in s:
            flt = float(s.split('*')[0]) * pi
        elif '-pi/' in s:
            flt = -pi / float(s.split('/')[1])
        elif 'pi/' in s:
            flt = pi / float(s.split('/')[1])
        elif '-pi' in s:
            flt = -pi
        elif 'pi' in s:
            flt = pi
        else:
            flt = float(s)
        return flt

    gatelist = []
    for i in range(qc.depth()):  # splits circuit into steps
        dag = circuit_to_dag(qc)
        layers = list(dag.multigraph_layers())
        n_remove = qc.depth() - i

        # the extra minus 1 since the last layer consists of output nodes (qubits and clbits).
        for layer in layers[- n_remove:]:

            for node in layer:
                if node.type == 'op':
                    dag.remove_op_node(node)

        new_qc = dag_to_circuit(dag)
        gatelist.append(new_qc.qasm())

    newgatelist = []  # function for splitting lines in gatelist. gatelist[n][k] gives nth depth, and kth element
    for i in range(qc.depth()):
        newgatelist.append(gatelist[i].split('\n'))
        del newgatelist[i][0]
        del newgatelist[i][0]
        del newgatelist[i][0]
        if len(qc.clbits) != 0:
            del newgatelist[i][
                0]  # deletes the first three or four lines of the qasm file which contain the filetype, ammount of qubits etc..
        del newgatelist[i][-1]

    for k in range(qc.depth() - 1):
        for i in range(len(newgatelist[qc.depth() - 2 - k])):  # näst sista element i lista, jobba sig neråt
            x = newgatelist[qc.depth() - 2 - k][i]  # näst sista listan, jobba sig neråt
            for j in range(len(newgatelist[qc.depth() - 1 - k])):
                if x == newgatelist[qc.depth() - 1 - k][j]:
                    del newgatelist[qc.depth() - 1 - k][j]
                    break

    a = newgatelist

    allgates = []  # lists that will contain the gates
    allargs = []
    allqubits = []

    for k in range(len(a)):
        # find gatetype by reading everything until a "(" or "q" is detected
        gates = []
        for i in range(len(a[k])):
            if a[k][i].__contains__('('):
                ind = a[k][i].index('(')
                gates.append(a[k][i][0:ind])
            else:
                ind = a[k][i].index('q')
                gates.append(a[k][i][0:ind - 1])

        # find argument by reading whatever is inside of the parenthesis
        args = []
        for i in range(len(a[k])):
            if "(" in a[k][i]:
                args.append(piDelivery(a[k][i][a[k][i].find("(") + 1:a[k][i].find(
                    ")")]))  # find item inside parenthesis, if no parenthesis current append 0
            else:
                args.append(0)

        # find which qubits are acted on in each step by finding reading what's inside the "[]". If it's a two qubit gate both are added
        qubits = []
        for i in range(len(a[k])):
            lastbit = a[k][i][-3:-2]
            result = (search(r"\[([A-Za-z0-9_]+)\]", a[k][i]))
            if result.group(1) == lastbit:
                qubits.append(int(lastbit))
            else:
                qubits.append([int(result.group(1)), int(lastbit)])

        allgates.append(gates)
        allargs.append(args)
        allqubits.append(qubits)

    for i in range(qc.depth()):
        for j in range(len(allgates[i])):
            if allgates[i][j] == "cz":
                allgates[i][j] = "CZ"
            if allgates[i][j] == "ry":
                allgates[i][j] = "PY"
            if allgates[i][j] == "rx":
                allgates[i][j] = "PX"
            if allgates[i][j] == "h":
                allgates[i][j] = "HD"
            if allgates[i][j] == "iswap":
                allgates[i][j] = "iSWAP"
            if allgates[i][j] == "rz":
                allgates[i][j] = "VPZ"
            if allgates[i][j] == "id":
                allgates[i][j] = "ID"

    toqnas = []
    for j in range(qc.depth()):
        toqnas.append([allgates[j], allqubits[j], allargs[j]])

    steps = []
    for stp in enumerate(toqnas):
        steps.append(gf.AlgStep(stp[1][0], stp[1][1], stp[1][2]))
    return steps



__all__ = ['mainAlgorithm']

from . import gateFuncs as gf
from . import mcSolving
from .anharmonicity import anharmonicity
from . import storeTimeDynamicsMain
from . import zzInteractionFunction as zz
from qutip import parfor
def mainAlgorithm(args):
    """
    Main algorithm.
    parallell for loop solver for the ntraj trajectories for the entire quantum algorithm.
    Input: in form of an args dictionary
    - steps = list of AlgStep objects
    - c_ops = list of collapse operators
    - psi0 = initial state (|00000 .. >)
    - Qblist = list of Qubit objects
    - t_max = list of maximum gate times for 1 and 2qb gates
    - ntraj = number of trajectories
    - zz_mat = matrix for the ZZ interaction, optional.
    - StoreTimeDynamics = boolean for option to store the time dynamics.
            Then also input on e_ops is needed.
            - e_ops_inp = [[e_op1, Tar_Con],
                            [e_op2, Tar_Con], ... ]
                        where e_op is a Qobj with the dimensions  ( qubit.level x qubit.level )
                        and Tar_Con is the target and control in case of 2qb gate, as before.
    Output: ntraj number of final states
    or if StoreTimeDynamics = True: also list of expectation values at each time and total time list
    """
    steps = args["steps"]
    c_ops = args["c_ops"]
    init_state = args["psi0"]
    Qblist = args["Qblist"]
    t_max = args["t_max"]
    ntraj = args["ntraj"]
    StoreTimeDynamics = args["StoreTimeDynamics"]

    psi0 = []
    for i in range(ntraj):
        psi0.append(init_state)

    try:
        zz_mat = args["zz_mat"]
        H0 = anharmonicity(Qblist) + zz.zzInteraction(Qblist, zz_mat)
    except: # if zz interactions not specified, we skip them
        H0 = anharmonicity(Qblist)

    if not StoreTimeDynamics:
        if c_ops != []:
            for i in range(0,len(steps)):
                physicalgates, virtualgates, tmax = gf.createGatesFromStep(steps[i], Qblist, t_max)
                Htd, tlist = gf.timeDepend(steps[i], physicalgates, tmax, Qblist)
                H = Htd + H0
                if max(tlist) >= 1e-11:  # To avoid integration error
                    psi0 = parfor(mcSolving.mcs, psi0, H=H, tlist=tlist, c_ops=c_ops)
                for vgate in virtualgates:
                    psi0= parfor(mcSolving.virtgate, psi0, vgate=vgate)
        else:
            for i in range(0,len(steps)):
                physicalgates, virtualgates, tmax = gf.createGatesFromStep(steps[i], Qblist, t_max)
                Htd, tlist = gf.timeDepend(steps[i], physicalgates, tmax, Qblist)
                H = Htd + H0
                if max(tlist) > 1e-11: # To avoid integration error
                    psi0 = parfor(mcSolving.mcs, psi0, H=H, tlist=tlist, c_ops=c_ops)
                for vgate in virtualgates:
                    psi0 = parfor(mcSolving.virtgate, psi0, vgate=vgate)
        return psi0
    if StoreTimeDynamics:
        psi, expectvals, tlist_tot = storeTimeDynamicsMain.mainAlgorithmExpectation(args)
        return psi, expectvals, tlist_tot
    



__all__ = ['Qubit', 'createPsi0']


from qutip import basis, tensor

class Qubit:
    """ Definition of Qubit object class
    Includes:
    - level = qubit levels (eg 2, 3 or 4)
    - noisert_vec = list of noise rates [relaxation, dephasing, excitation]
    - anharm = anharmonicity value for the qubit """
    def __init__(self, level, noisert_vec, anharm):
        self.level = level
        self.noisert_vec = noisert_vec
        self.anharm = anharm


def createPsi0(Qblist, init_state):
    """ Creates initial state with correct dimensions
    Input:
    - Qblist = list of Qubit objects
    - init_state = int of initial state value (often =0)"""
    psi0 = [basis(Qb.level, init_state) for Qb in Qblist]
    return tensor(psi0)




# Importing all functions available, gives: import qnas as x -> x.function()
from .help import *
from .anharmonicity import *
from .collapseOperatorFunction import *
from .envelopeFunction import *
from .gateFuncs import *
from .gateLib import *
from .help import *
from .mainAlgorithm import *
from .mcSolving import *
from .openqasmInterpreter import *
from .qubitClass import *
from .readData import *
from .solve import *
from .zzInteractionFunction import *
from .storeTimeDynamicsMain import *




__all__ = ['tensorifyExpectationOperator', 'createExpectOpsList', 'mainAlgorithmExpectation']

from qutip import qeye, tensor, parfor, Qobj, expect
from numpy import pi, array, concatenate, transpose, delete, reshape, append, mean
from . import gateLib
from sys import exit
from . import gateFuncs as gf
from . import mcSolving
from .anharmonicity import anharmonicity
from . import zzInteractionFunction as zz


def tensorifyExpectationOperator(Qblist, Tar_Con, Gate):
    """
    Tensors out gates for which to calculate the expectation values.
    Inputs:
    - Qblist (list of Qubit objects)
    - Tar_Con = int for 1qb gates, and list [i,j] for 2qb gates. Which gate to target
    - Gate = Qobj of dimensions corresponding to the targeted qubit levels. (N x N) if qubit.level = N
    Output: Tensored gate with correct dimension
    """
    gateList = [qeye(Qb.level) for Qb in Qblist]
    if type(Tar_Con) == int:  # 1qb gate
        gateList[Tar_Con] = Gate
        if Gate.shape[0] != Qblist[Tar_Con].level:
            print('Error: Shape of e_ops doesn\'t match the targeted qubit level. \n See \'storeTimeDynamics\' for more info')
            exit(1)
        return tensor(gateList)
    elif len(Tar_Con) == 2: # 2qb gate
        if Gate.shape[0] != Qblist[Tar_Con[0]].level*Qblist[Tar_Con[1]].level:
            print('Error: Shape of e_ops doesn\'t match the targeted qubits levels. \n See \'storeTimeDynamics\' for more info')
            exit(1)
        del (gateList[max(Tar_Con)])  # Make room for the gate
        del (gateList[min(Tar_Con)])  # Make room for the gate
        return gateLib.gate_expand_2toN(Gate, len(Qblist), gateList, Tar_Con[1], Tar_Con[0])
    else:
        print('Error: QnAS only handles 1 or 2 qubit operators as e_ops. \n See \'ExpectationValues\' for more info')
        exit(1)  # Stops the program with the same error code as above

def createExpectOpsList(Qblist, e_ops_inp):
    """
    Input given on the form:
    e_ops_inp = [[e_op1, Tar_Con],
                 [e_op2, Tar_Con], ... ]
    where e_op is a Qobj with the dimensions  ( qubit.level x qubit.level )
    and Tar_Con is the target and control in case of 2qb gate, as before.
    output: List of expectation value operators (e_ops)
    """
    e_ops=[]
    for i in range(len(e_ops_inp)):
        e_ops.append(tensorifyExpectationOperator(Qblist, e_ops_inp[i][1], e_ops_inp[i][0]))
    return e_ops


def mainAlgorithmExpectation(args):
    """
    mainAlgorithm modified to incorporate the StoreTimeDynamics option, by giving output:
    -final state
    -tlist for the entire algorithm
    -expectation values for a user specified set of e_ops, at each time step in the tlist.
    See mainAlgorithm.py for more information
    """
    steps = args["steps"]
    c_ops = args["c_ops"]
    init_state = args["psi0"] # maybe we should change the input so we can write args["init_state"] instead
    Qblist = args["Qblist"]
    t_max = args["t_max"]
    ntraj = args["ntraj"]
    tlist_tot = [0]
    e_ops_inp = args["e_ops_inp"]
    allStates = array([])  # a list where all states are saved

    psi0 = []
    for i in range(ntraj):
        psi0.append(init_state)

    e_ops = createExpectOpsList(Qblist, e_ops_inp) # Construct the e_ops array from e_ops_input

    try:
        zz_mat = args["zz_mat"]
        H0 = anharmonicity(Qblist) + zz.zzInteraction(Qblist, zz_mat)
    except:  # if zz interactions not specified, we skip them
        H0 = anharmonicity(Qblist)

    if c_ops != []:
        for i in range(0, len(steps)):  # all steps
            physicalgates, virtualgates, tmax = gf.createGatesFromStep(steps[i], Qblist, t_max)
            Htd, tlist = gf.timeDepend(steps[i], physicalgates, tmax, Qblist)
            H = Htd + H0
            # Create tlist for the entire process
            OnlyVirtualGates = True
            for j in range(len(steps[i].name)): # Deals with steps with multiple gates in them
                if gateLib.isPhysicalGate(steps[i], j) or gateLib.isTwoQubitGate(steps[i], j) or steps[i].name[j] in ["HD"]:
                    OnlyVirtualGates = False
            if OnlyVirtualGates or max(steps[i].angle) < 0.01:
                tlist_shifted = []
            else:
                tlist_shifted = tlist + tlist_tot[-1]  # Shifting the tlist to start where previous ends.
            tlist_tot = concatenate((tlist_tot, tlist_shifted))
            # TODO : KOLLA ATT DET HÄR VERKLIGEN BLIR RÄTT ^
            # Jag tror att det ska funka som det är nu. / Ed
            if max(tlist) >= 1e-11:
                allStates = append(allStates, transpose(parfor(mcSolving.mcsTimeDynamics, psi0, H=H, tlist=tlist, c_ops=c_ops)))
                psi0 = allStates[-ntraj:]
            for vgate in virtualgates:
                psi0 = parfor(mcSolving.virtgate, psi0, vgate=vgate)
        tlist_tot = delete(tlist_tot, 0)  # We get double zero in the beginning since tlist_tot = [0] initially
        allStates = reshape(allStates, (len(tlist_tot), ntraj))
        # TODO: Koolla på detta igen ^
    else:
        psi0 = psi0[0]#we only need one simulation if there is no noise
        allStates = []
        for i in range(0, len(steps)):  # all steps
            physicalgates, virtualgates, tmax = gf.createGatesFromStep(steps[i], Qblist, t_max)
            Htd, tlist = gf.timeDepend(steps[i], physicalgates, tmax, Qblist)
            H = Htd + H0
            # Create tlist for the entire process
            OnlyVirtualGates = True
            for j in range(len(steps[i].name)):  # Deals with steps with multiple gates in them
                if gateLib.isPhysicalGate(steps[i], j) or gateLib.isTwoQubitGate(steps[i], j) or steps[i].name[j] in [
                    "HD"]:
                    OnlyVirtualGates = False
            if OnlyVirtualGates or max(steps[i].angle) < 0.01:
                tlist_shifted = []
            else:
                tlist_shifted = tlist + tlist_tot[-1]  # Shifting the tlist to start where previous ends.
            tlist_tot = concatenate((tlist_tot, tlist_shifted))
            if max(tlist) > 1e-11:
                allStates = allStates + ((((mcSolving.mcsTimeDynamics( psi0, H=H, tlist=tlist, c_ops=c_ops)))))
                psi0 = allStates[-1]
            for vgate in virtualgates:
                psi0 = mcSolving.virtgate( psi0, vgate=vgate)
        tlist_tot = delete(tlist_tot, 0)  # We get double zero in the beginning since tlist_tot = [0] initially

       
    """CALCULATE EXPECTATION VALUES FROM ALL STATES HERE """
    if type(e_ops) == list and type(e_ops[0] == Qobj):
        expectvals = [array([mean(expect(e, parallelStates)) for parallelStates in allStates]) for e in e_ops]
    else:
        raise Exception("e_ops needs to be a list of Qobj")

    return psi0, expectvals, tlist_tot
#    return finalstate, expectvals, [tlist_tot]
"""Added parenthesis to tlist_tot so that tlist_tot will be an 2D array gives a reasonable format for plotting 
(I got an error if I didn't have it, but maybe we can avoid it in the future) //Albin"""




import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name='qnas',  # should match the package folder
    version='0.0.1',  # important for updates
    license='MIT',  # should match your chosen license
    description='QnAS package',
    long_description=long_description,  # loads your README.md
    long_description_content_type="text/markdown",  # README.md is of type 'markdown'
    author='Isak Wikman, Edvin Martinson, Axel Blom, Albin Edenmyr, Didrik Palmqvist, Ludvig Nordqvist',
    author_email='isakwi@student.chalmers.se, edvmar@student.chalmers.se, bloaxel@student.chalmers.se, edenmyr@student.chalmers.se, didrikp@student.chalmers.se, ludnor@student.chalmers.se',
    url='https://github.com/isakwi/Kandidatarbete/tree/main/qnas',  # Skapa en ny git för detta?!
    install_requires=['numpy', 'math', 'qutip', 'qiskit', 'pandas', 'sys', 're'],  # list all packages that your package uses
    keywords=["pypi", "qnas", "algorithm-simulator", "quantum", "quantum-algorithm-simulator", "algorithm","noise","simulator"],  # descriptive meta-data
    classifiers=[  # https://pypi.org/classifiers
        'Development Status :: 3 - Alpha',
        'Topic :: Software Development :: Documentation',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],

    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src"),
    python_requires=">=3.9",
)


__all__ = ['solve']

from . import qubitClass as qbc
from numpy import pi
from . import readData as rd
from . import collapseOperatorFunction as co
from . import mainAlgorithm
from . import openqasmInterpreter as opq
def solve(Qbfile = None, circuit = None, zz_int = None, ntraj=500, tmax=None, store_time_dynamics = False, e_ops=None):
    """
    The main solver function. Basically a user calls this function and everything else is automatic
    :param Qbfile: File that holds qubit parameters. Default - 3 levels, No noises, anharmonicity -225e6*2*pi
    :param circuit: Qiskit QuantumCircuit object to specify quantum circuit. Can't run without
    :param zz_int: symmetrical square (n x n) matrix that describes interaction between qubits
    :param ntraj: number of trajectories for the Monte Carlo solver. Default - 500
    :param tmax: Max time for 1qb-gate and 2qb-gate ~ [t_1qb, t_2qb]. Default - [20e-9, 200e-9]
    :param store_time_dynamics: True/False value to store time dynamics. Default - False
    :param e_ops: Expectation value operators for store_time_dynamics. Given as [[e_op1, Tar_Con],[e_op2, Tar_Con], ...]
    :return: if store_time_dynamics is True: Returns ntraj many final states and two lists with exp values
            and corresponding times. Else: ntraj many final states
    """
    if e_ops is None:
        e_ops = []
    if circuit is None:
        print("You didn't enter an circuit circuit. QnAS will now exit?\t")
        return

    try:
        steps = opq.qasmToQnas(circuit)
    except:
        print(f"Couldn't read the circuit file! Check that the circuit, {circuit}, is correct and that the circuit is"
              "constructed correctly. \nQnAS.solve() will now exit")
        return

    # Check n
    # Think this will work
    try:
        n = circuit.num_qubits
    except:
        print("Couldn't determine number of qubits from circuit."
              "\nQnAS.solve() will now exit")

    if not 1 <= n <= 15:
        print("Invalid value for number of qubits! Must be between 1 and 15! \nQnAS.solve() will now exit")
        return

    # Read qubit parameters
    if Qbfile is None:
        print("You didn't specify a file for the qubit parameters! \n"
              "QnAS assumes you have 3 levels, no noises and sets the anharmonicity to"
              " -225e6*2*pi for all qubits")
        Qblist = [qbc.Qubit(3, [0,0,0], -225e6 * 2 * pi) for i in range(n)]
    else:
        try:
            relax, depha, therma, anharm, levels = rd.readFile(Qbfile, n)
            Qblist = []
        except:
            print(f"Couldn't find file {Qbfile}. QnAS.solve() will now exit")
            return
        for i in range(0, n):  # Creates list with all qubits, for now the desig and init_vec are empty
            anharm[i] = -2 * pi * abs(anharm[i]) * 1e6  # Convert linear frequency to angular (input seems to usually be linear)
            Qblist.append(qbc.Qubit(levels[i], [relax[i], depha[i], therma[i]], anharm[i]))

    if not 1 <= ntraj <= 100000:
        print("Invalid value for ntraj! Must be a positive integer with an upper limit of 100000 (to not kill your"
              " computer). If you leave it unspecified, ntraj will automatically be 500. QnAS.solve() will now exit")
        return

    if tmax is None:
        print("No input for tmax! QnAS will use 1qb gate time: 20ns and 2qb gate time: 200ns!")
        tmax = [20e-9, 200e-9]

    if not (type(tmax) == list and len(tmax) == 2):
        print("Invalid input for tmax! tmax must be a list of the form [1qb, 2qb] with "
              "max gate times for 1-qubit-gates and 2-qubit-gates given in seconds!"
              " QnAS.solve() will now exit")
        return
    if not (1e-12 < tmax[0] < 1e12 or 1e-12 < tmax[1] < 1e12):
        print("Invalid values for at least one of the max gate times. Limits are 1e-12 < t < 1e12. "
              " QnAS.solve() will now exit")
        return
    if type(store_time_dynamics) != bool:
        print("store_time_dynamics must be a boolean (True/False). "
              "QnAS.solve() will now exit")
        return
    if store_time_dynamics == True:  # Only need to check e_ops if store_td is true
        if e_ops is None:
            print("You didn't enter any e_ops, no need to save time dynamics!")
            store_time_dynamics = False
        if type(e_ops) != list:
            print("Wrong input type of e_ops! Input given on the form:\n"
                  "e_ops_inp = [[e_op1, Tar_Con],[e_op2, Tar_Con], ... ]\n"
                  "where e_op is a Qobj with the dimensions  ( qubit.level x qubit.level )\n"
                  "and Tar_Con is the target and control in case of 2qb gate, as before.\n"
                  "QnAS.solve() will now exit")
            return
        for e_op in e_ops:
            if type(e_op) != list:
                print("Wrong input type of e_ops! Input given on the form:\n"
                      "e_ops_inp = [[e_op1, Tar_Con],[e_op2, Tar_Con], ... ]\n"
                      "where e_op is a Qobj with the dimensions  ( qubit.level x qubit.level )\n"
                      "and Tar_Con is the target and control in case of 2qb gate, as before.\n"
                      "QnAS.solve() will now exit")
                return
            #if type(e_op[0]) != Qobj:  # Would like to add something like this but don't know how

    if zz_int is not None:
        if type(zz_int) != list:
            print("Wrong format for the interaction matrix! Should be given as [[x1, x2, ... , xn], [...]]"
                  "\nfor n qubits! QnAS.solve() will now exit")
            return
        if not n <= len(zz_int) <= 15:
            print("The matrix can't be smaller than the number of qubits (or larger than 15)! \n"
                  "QnAS.solve() will now exit")
            return
        for row in zz_int:
            if not n <= len(row) <= 15:
                print("The matrix can't be smaller than the number of qubits (or larger than 15)!\n "
                      "QnAS.solve() will now exit")
                return

    psi0 = qbc.createPsi0(Qblist, 0)
    c_ops = co.createCollapseOperators(Qblist)

    if zz_int is None:
        args = {"steps" : steps, "c_ops" : c_ops, "psi0" : psi0, "Qblist": Qblist, "t_max": tmax, "ntraj" : ntraj,
                "StoreTimeDynamics": store_time_dynamics, "e_ops_inp": e_ops}

    else:
        args = {"steps": steps, "c_ops": c_ops, "psi0": psi0, "Qblist": Qblist, "t_max": tmax, "ntraj": ntraj,
                "StoreTimeDynamics": store_time_dynamics, "e_ops_inp": e_ops, "zz_mat": zz_int}

    return mainAlgorithm(args)



__all__ = ['help']

def help():
    """
    User calls this function to recieve information about how to use QnAS
    """

    print("\n\nHello, welcome to QnAS - Quantum noisy Algorithm simulator!\n"
          "Here comes some help on how to get started :)\n\n"
          "The program is made to be able to simulate quantum algorithms, with noise,"
          " from the gates available at WACQT. All gates available are:\n"
          "\"PX\" Pauli x gate\n"
          "\"PY\" Pauli y gate\n"
          "\"VPZ\" Pauli z gate but this one is virtual, takes no time, hence the V in in VPZ\n"
          "\"HD\" Hadamard gate\n"
          "\"CZ\" Controlled z gate\n"
          "\"iSWAP\" iSWAP gate\n"
          "Make sure these are the only gates you have included in your circuit to create the algorithm!\n"
          "\nDescription of functions: \n\n"
          "help()\n"
          "\t:You already called this function so you probably know that "
          "it just helps you get started!\n\n"
          "solve(Qbfile, circuit, n, ntraj, tmax, store_time_dynamics)\n"
          "\t:param Qbfile: Filepath that holds qubit parameters. Default - 3 levels, No noises, anharmonicity -225e6*2*pi\n"
          "\t:param circuit: Filepath that specifies circuit file. Default - Asks user to specify gates manually or not run\n"
          "\t:param zz_int: symmetrical square (n x n) matrix that describes interaction between qubits.\n"
          "\t:param ntraj: number of trajectories for the Monte Carlo solver. Default - 500\n"
          "\t:param tmax: Max time for 1qb-gate and 2qb-gate ~ [t_1qb, t_2qb]. Default - [20e-9, 200e-9]\n"
          "\t:param store_time_dynamics: True/False value to store time dynamics. Default - False\n"
          "\t:param e_ops: Expectation value operators for store_time_dynamics. Given as [[e_op1, Tar_Con],[e_op2, Tar_Con], ...]\n"
          "\t:return: if store_time_dynamics is True: Returns ntraj many final states and two lists with exp values "
          "and corresponding times. Else: ntraj many final states\n\n"
          "The qubit parameter file needs to be located in the folder where you are running your program.\n"
          "Here is an example qubit file structure (OBS! .csv file!):\n"
          "\tQubit;Relaxation;Dephasing;Thermal;Anharmonicity(MHz);Levels\n"
          "\t01;0.00;0.00;0.00;-229;3\n"
          "\t02;0.01;0.00;0.00;-225;3\n"
          "\t03;0.01;0.00;0.00;200;3\n"
          "\t04;0.04;0.00;0.00;200;3\n"
          "\t05;0.05;0.00;0.00;200;2\n"
          "\t06;0.06;0.00;0.00;200;2\n"
          "\t07;0.07;0.00;0.00;200;2\n"
          "\t08;0.08;0.00;0.00;200;2\n"
          "\t09;0.09;0.00;0.00;200;2\n"
          "\t10;0.10;0.00;0.00;200;2\n"
          "\t11;0.11;0.00;0.00;200;2\n"
          "\t12;0.12;0.00;0.00;200;2\n"
          "\t13;0.02;0.00;0.00;200;2\n"
          "\t14;0.02;0.00;0.00;200;2\n"
          "\t15;0.02;0.00;0.00;200;2\n\n"
          "Do you want to donate to QnAS development to keep the project going forward?\n"
          "Swish to +46 70 603 61 27, thanks!")
